/* libcm - A library with a xlib wrapper and a gl based scene graph
 * Copyright (C) 2005, 2006  Red Hat, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 *
 * Author: Soren Sandmann (sandmann@redhat.com)
 */
#include "drawable-node.h"
#include <string.h>
#include "wsint.h"
#include "state.h"
#include <math.h>
#include "stacker.h"

#include <X11/Xlib.h>
#include <X11/extensions/Xfixes.h>

#if G_BYTE_ORDER == G_LITTLE_ENDIAN
#   define PIXEL_TYPE GL_UNSIGNED_INT_8_8_8_8_REV
#else
#   define PIXEL_TYPE GL_UNSIGNED_INT_8_8_8_8
#endif

G_DEFINE_TYPE (CmDrawableNode, cm_drawable_node, CM_TYPE_NODE);

static void drawable_node_render (CmNode *node,
				  CmState *state);
static void drawable_node_compute_extents (CmNode  *node,
					   Extents *extents);

static void
cm_drawable_node_finalize (GObject *object)
{
    CmDrawableNode *node = CM_DRAWABLE_NODE (object);
    WsDisplay *display = WS_RESOURCE (node->drawable)->display;
    
    ws_display_begin_error_trap (display);
    
    if (node->pix_texture->pixmap)
	ws_pixmap_set_damage_callback (node->pix_texture->pixmap, NULL, NULL);
    
    if (node->pix_texture)
	g_object_unref (node->pix_texture);
    
    if (node->configure_connection)
	g_signal_handler_disconnect (node->drawable,
				     node->configure_connection);
    
    ws_display_end_error_trap (display);
    
    /* FIXME: remove this when Ws* has been gobjectified */
    G_OBJECT_CLASS (cm_drawable_node_parent_class)->finalize (object);
}

static void
cm_drawable_node_class_init (CmDrawableNodeClass *class)
{
    GObjectClass *object_class = G_OBJECT_CLASS (class);
    CmNodeClass *node_class = CM_NODE_CLASS (class);
    
    object_class->finalize = cm_drawable_node_finalize;
    
    node_class->render = drawable_node_render;
    node_class->compute_extents = drawable_node_compute_extents;
}

static void
cm_drawable_node_init (CmDrawableNode *drawable_node)
{
    CmStacker *stacker;
    
    drawable_node->grid =
	cm_grid_new (0, 0, 20, 20, 32, 32);
    drawable_node->pix_texture =
	cm_pix_texture_new (NULL,
			    CM_NODE (drawable_node->grid));
    stacker = cm_stacker_new ();
    drawable_node->shadow = cm_shadow_new (20, 20, 32, 32);
    cm_stacker_add_child (stacker, CM_NODE (drawable_node->shadow));
    cm_stacker_add_child (stacker, CM_NODE (drawable_node->pix_texture));
    drawable_node->deform =
	cm_deform_new (CM_NODE (stacker), 0, 0, 32, 32);
}

static void
queue_paint (CmDrawableNode *node)
{
    cm_node_queue_repaint (CM_NODE (node));
}

static void
print_matrix (const char *header, GLdouble m[16])
{
    int i, j;
    
    g_print ("%s:\n", header);
    
    for (i = 0; i < 16; ++i)
    {
	g_print ("%f ", m[i]);
	
	if (++j == 4)
	{
	    g_print ("\n");
	    j = 0;
	}
    }
}

static void
print_rectangle (const char *header, WsRectangle *rect)
{
    g_print ("%s\n", header);
    
    g_print ("%d %d %d %d\n", rect->x, rect->y, rect->width, rect->height);
}

static void
print_coord (int x, int y, int z)
{
    GLdouble model[16];
    GLdouble proj[16];
    GLint    view[4];
    double   wx, wy, wz;
    
    glGetDoublev (GL_MODELVIEW_MATRIX, model);
    glGetDoublev (GL_PROJECTION_MATRIX, proj);
    glGetIntegerv (GL_VIEWPORT, view);
    
    gluProject (x, y, z,
		model, proj, view,
		&wx, &wy, &wz);
    
#if 0
    print_matrix ("model", model);
    print_matrix ("proj",  proj);
    g_print ("viewport: %d %d %d %d\n", view[0], view[1], view[2], view[3]);
    
    g_print ("projected: %d %d %d  =>   %f %f %f\n", x, y, z, wx, wy, wz);
#endif
}

static void
update_geometry (CmDrawableNode *node)
{
    WsRectangle geometry;
    
    if (!ws_drawable_query_geometry (WS_DRAWABLE (node->drawable), &geometry))
    {
	return;
    }
    
    node->real_x = geometry.x;
    node->real_y = geometry.y;
    node->real_width = geometry.width;
    node->real_height = geometry.height;
}

static void
drawable_node_vertex (gdouble *x,
		      gdouble *y,
		      gdouble *z,
		      gdouble  u,
		      gdouble  v,
		      gpointer data)
    
{
    CmDrawableNode *dnode = data;
    
    glTexCoord2f (u * dnode->real_width,
		  v * dnode->real_height);
}

static void
set_geometries (CmDrawableNode *dnode)
{
    WsRectangle geometry;
    WsRegion *reg = NULL;

    if (dnode->saved_shape)
    {
	reg = ws_region_copy (dnode->saved_shape);
    }
    else
    {
	reg =
	    ws_window_get_output_shape (WS_WINDOW (dnode->drawable));
    }
	
    ws_region_offset (reg, dnode->real_x, dnode->real_y);

    cm_grid_set_region (dnode->grid, reg);
    
    ws_region_get_clipbox (reg, &geometry);
    
#if 0
    print_rectangle ("drawing geometry", &geometry);
#endif
    
    cm_deform_set_geometry (dnode->deform,
			    geometry.x,
			    geometry.y,
			    geometry.width,
			    geometry.height);
    cm_shadow_set_geometry (dnode->shadow,
			    geometry.x,
			    geometry.y,
			    geometry.width,
			    geometry.height);
    
    ws_region_destroy (reg);	
}

static void
drawable_node_render (CmNode *node,
		      CmState *state)
{
    WsScreen *screen;
    CmDrawableNode *dnode = CM_DRAWABLE_NODE (node);
    
    if (!dnode->viewable)
    {
#if 0
	g_print ("not rendering %lx since it is not viewable\n",
		 WS_RESOURCE_XID (dnode->drawable));
#endif
	return;
    }
    
    if (!dnode->pix_texture)
    {
#if 0
	g_print ("not rendering %lx since it doesn't have a pix_texture\n",
		 WS_RESOURCE_XID (dnode->drawable));
#endif
	return;
    }
    
    if (!dnode->pix_texture->pixmap)
    {
#if 0
	g_print ("not rendering %lx since it doesn't have a pixmap\n",
		 WS_RESOURCE_XID (dnode->drawable));
#endif
	return;
    }
    
#if 0
    g_print ("rendering %lx\n", WS_RESOURCE_XID (dnode->drawable));
#endif
    
    screen = ws_drawable_query_screen (
	WS_DRAWABLE (dnode->pix_texture->pixmap));
    
    if (dnode->real_width == -1)
	update_geometry (dnode);
    
    glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
    
    glPushMatrix();
    
    gluOrtho2D (0,
		ws_screen_get_width (screen),
		ws_screen_get_height (screen), 0);
    
    /* Empirically determined constant that gets rid of some fuzzyness
     */
    glTranslatef (-0.125, -0.125, 0);
    
#if 0
    g_print ("setting geometry: %d %d %d %d\n",
	     dnode->real_x, dnode->real_y,
	     dnode->real_width, dnode->real_height);
#endif
    
    set_geometries (dnode);
    
    glColor4f (1.0, 1.0, 1.0, dnode->alpha);
    glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glEnable (GL_BLEND);
    glBlendFunc (dnode->alpha < 0.99 ? GL_SRC_ALPHA :
		 GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
    
#if 0
    cm_node_render (CM_NODE (dnode->pix_texture), state);
#endif
    
#if 0
    glTranslatef (20 * g_random_double(), 20 * g_random_double(), 0.0);
#endif
    
    cm_node_render (CM_NODE (dnode->deform), state);
    
    glPopMatrix();
    
    glBindTexture (GL_TEXTURE_RECTANGLE_ARB, 0);
}

static void
drawable_node_compute_extents (CmNode  *node,
			       Extents *extents)
{
    extents->x = 0.0;
    extents->y = 0.0;
    extents->z = 0.0;
    extents->width = 1.0;
    extents->height = 1.0;
    extents->depth = 1.0;
}

static void
on_damage (WsPixmap *pixmap, gpointer data)
{
    CmDrawableNode *node = data;
    
    queue_paint (node);
}

static void
refresh_pixmap (CmDrawableNode *node)
{
    WsWindow *window;
    WsDisplay *display = WS_RESOURCE (node->drawable)->display;
    WsPixmap *pixmap;
    
    if (node->pix_texture->pixmap && !node->pix_texture->pixmap->do_updates)
    {
	g_print ("not refreshing\n");
	return;
    }

#if 0
    g_print ("refresh\n");
#endif
    
    window = WS_WINDOW (node->drawable);
    
    ws_display_begin_error_trap (display);
    
    pixmap = NULL;
    
    if (ws_window_query_mapped (window)		&&
	!ws_window_query_input_only (window))
    {
	pixmap = ws_window_name_pixmap (window);
    }
    
    if (ws_display_end_error_trap_with_return (display))
    {
	ws_display_begin_error_trap (display);
	
	if (pixmap)
	{
	    g_object_unref (pixmap);
	    
	    pixmap = NULL;
	}
	
	ws_display_end_error_trap (display);
    }

    WsRectangle pixgeo;

    if (pixmap)
    {
	ws_drawable_query_geometry (pixmap, &pixgeo);

	if (node->saved_shape)
	    ws_region_destroy (node->saved_shape);
	
	ws_display_begin_error_trap (display);
	
	node->saved_shape =
	    ws_window_get_output_shape (WS_WINDOW (node->drawable));

	WsRectangle window_geo;

	ws_drawable_query_geometry (WS_WINDOW (node->drawable), &window_geo);

	WsRectangle pixmap_geo;

	ws_drawable_query_geometry (pixmap, &pixmap_geo);

#if 0
	if (window_geo.width  != pixmap_geo.width)
	    g_print ("wrong width of window\n");
#endif
	
#if 0
	ws_region_get_clipbox (node->saved_shape, &saved_box);

	print_rectangle ("saving box: ", &saved_box);
#endif
	
	ws_display_end_error_trap (display);
    }
#if 0
    else
	g_print ("nod pixmap\n");
#endif
    
    cm_pix_texture_set_pixmap (node->pix_texture, pixmap);
    
    if (pixmap)
    {
	ws_pixmap_set_damage_callback (pixmap, on_damage, node);
	
	ws_display_begin_error_trap (display);
	
	g_object_unref (pixmap);
	
	ws_display_end_error_trap (display);
    }
}

static void
on_configure (WsWindow *window,
	      WsConfigureEvent *event,
	      gpointer data)
{
    CmDrawableNode *node = data;
    
    if (node->pix_texture->pixmap &&
	!ws_pixmap_get_updates (node->pix_texture->pixmap))
    {
	return;
    }
    
    // g_print ("configure\n");
    if (event->width != node->real_width ||
	event->height != node->real_height)
    {
	refresh_pixmap (node);
    }
    
    node->real_x = event->x;
    node->real_y = event->y;
    node->real_width = event->width;
    node->real_height = event->height;
    
    queue_paint (node);
}

CmDrawableNode *
cm_drawable_node_new (WsDrawable *drawable)
{
    CmDrawableNode *node;
    WsDisplay *display;
    
    g_return_val_if_fail (drawable != NULL, NULL);
    
    display = WS_RESOURCE (drawable)->display;
    
    node = g_object_new (CM_TYPE_DRAWABLE_NODE, NULL);
    
#if 0
    g_print ("creating drawable node %p containing %lx\n",
	     node, WS_RESOURCE_XID (drawable));
#endif
    
    node->drawable = drawable;
    
    node->configure_connection =
	g_signal_connect (G_OBJECT (drawable),
			  "configure_event",
			  G_CALLBACK (on_configure), node);
    
    node->viewable = TRUE;
    
    node->timer = g_timer_new ();
    
    node->real_x = -1;
    node->real_y = -1;
    node->real_width = -1;
    node->real_height = -1;
    
    ws_display_init_composite (ws_drawable_get_display (drawable));
    ws_display_init_damage (ws_drawable_get_display (drawable));
    
    node->alpha = 1.0;
    
    if (WS_IS_WINDOW (drawable))
    {
	refresh_pixmap (node);
    }
    
    return node;
}

void
cm_drawable_node_set_viewable (CmDrawableNode *node,
			       gboolean	  viewable)
{
    node->viewable = viewable;
    
    cm_node_queue_repaint ((CmNode *)node);
}

void
cm_drawable_node_update_pixmap (CmDrawableNode *node)
{
#if 0
    g_print ("refreshing pixmap because app told us to\n");
#endif
    refresh_pixmap (node);
}

void
cm_drawable_node_set_geometry (CmDrawableNode *node,
			       double x,
			       double y,
			       double width,
			       double height)
{
    cm_deform_set_deform (node->deform, cm_user_geometry_deform, node);
    
    node->user_x = x;
    node->user_y = y;
    node->user_width = width;
    node->user_height = height;
    
    queue_paint (node);
}

void
cm_drawable_node_set_patch (CmDrawableNode *node,
			    CmPoint points[][4])
{
    g_return_if_fail (CM_IS_DRAWABLE_NODE (node));
    
    cm_deform_set_deform (node->deform, cm_patch_deform, node);
    
    memcpy (node->patch_points, points, sizeof node->patch_points);
    
    queue_paint (node);
}

void
cm_drawable_node_unset_geometry (CmDrawableNode *node)
{
    cm_deform_set_deform (node->deform, cm_identity_deform, NULL);
    
    queue_paint (node);
}

void
cm_drawable_node_set_updates (CmDrawableNode *node,
			      gboolean        updates)
{
    if (node->pix_texture->pixmap->do_updates == !!updates)
	return;

    g_print ("drawable node updates %d\n", updates);
    
    if (!updates && node->drawable)
    {
	WsDisplay *display = WS_RESOURCE (node->drawable)->display;
	
    }
    
    if (node->pix_texture->pixmap)
	ws_pixmap_set_updates (node->pix_texture->pixmap, updates);

    if (updates)
    {
	WsRectangle rect;

	ws_drawable_query_geometry (node->drawable, &rect);

	node->real_x = rect.x;
	node->real_y = rect.y;
	node->real_width = rect.width;
	node->real_height = rect.height;
    }
    
    refresh_pixmap (node);
}

void
cm_drawable_node_set_alpha (CmDrawableNode *node,
			    double alpha)
{
    node->alpha = alpha;
    
    queue_paint (node);
}
